Use Windows.pkg
Use DFClient.pkg
Use dfTreeVw.pkg
Use cProcessStatusPanel.pkg
Use dfSpnFrm.pkg
Use tNameData.pkg

Define C_SampleSize for 60000                   // This constant defines the total size of the possible set of names.
Define C_TestDataIncrementSize for 100          // This means we fetch 100 new names from the test data each time we repeat the test
Define C_TestDataFileName for "TestData.txt"    // Name of text file used to store our test data

// GetTickCount:

// Retrieves the number of milliseconds that have elapsed since Windows was started.
External_Function GetTickCount "GetTickCount" Kernel32.dll Returns Integer


// Sleep:

// Suspends the execution of the current thread until the time-out interval elapses.
External_Function Sleep "Sleep" Kernel32.dll DWord dwMilliseconds Returns Integer


// TimeBeginPeriod:

// Specifies the minimmum resolution in milliseconds for periodic timers.
External_Function TimeBeginPeriod "timeBeginPeriod" Winmm.dll UInteger uPeriod Returns Integer

// TimeEndPeriod:

// Cancels the previous call to TimeBeginPeriod.
External_Function TimeEndPeriod "timeEndPeriod" Winmm.dll UInteger uPeriod Returns Integer



Activate_View Activate_oLookupNames for oLookupNames
Object oLookupNames is a dbView
    Set Size to 375 507
    Set Location to 2 2
    Set pbAutoActivate to True
    
    // pTestData: This array stores the total set of possible names that can be sent to our test. This data is used to
    // feed our test simulator. Normally the total set would not be know to our application and would be passed to it from
    // some external source.
    Property String[] pTestData
    Property Integer piTestDataOffset 0
    
    // pNameDataCache: This is a cache of all names looked up so far stored as tNameData structs
    Property tNameData[] pNameDataCache

    // pNamesCache: A cache of searched names
    Property String[] pNamesCache
    
    // pDatesCache: A cache of all dates associated with the members of pNamesCache.
    Property Date[] pDatesCache

    // pSortedNamesCache: A sorted cache of searched names
    Property String[] pSortedNamesCache
    
    // pSortedDatesCache: A sorted cache of all dates associated with the members of pSortedNamesCache.
    Property Date[] pSortedDatesCache

    Set Border_Style to Border_Thick

    Object oSizeOfTest_grp is a RadioGroup
        Property Boolean pbFirstTime True
        
        Set Location to 5 6
        Set Size to 68 77
        Set Label to "Size of Test"
    
        Object oSize1 is a Radio
            Set Label to "1000 Names"
            Set Size to 10 61
            Set Location to 15 7
        End_Object
    
        Object oSize2 is a Radio
            Set Label to "10,000 Names"
            Set Size to 10 61
            Set Location to 33 7
        End_Object

        Object oSize3 is a Radio
            Set Label to "50,000 Names"
            Set Size to 10 61
            Set Location to 50 7
        End_Object


        // Notify_Select_State:

        // Augmented to clear the name data caches.        
        Procedure Notify_Select_State Integer NewId Integer OldId
            String[] EmptyNames
            Date[] EmptyDates
            tNameData[] EmptyNameData
            Forward Send Notify_Select_State NewId OldId
            
            If (not(pbFirstTime(Self))) Begin
                Set pNameDataCache to EmptyNameData
                Set pNamesCache to EmptyNames
                Set pDatesCache to EmptyDates
                Set pSortedNamesCache to EmptyNames
                Set pSortedDatesCache to EmptyDates
                Set Value of oTestCount1_frm to 0
                Set Value of oTestCount2_frm to 0
                Set Value of oTestCount3_frm to 0
                Set Value of oTestCount4_frm to 0
                Set Value of oTestCount5_frm to 0
            End
        End_Procedure  // Notify_Select_State
        

        // SizeOfTest:
        
        // Returns the selected number of names to use in our test.
        Function SizeOfTest Returns Integer
            Integer iSelectedSize
            
            Get Current_Radio to iSelectedSize
            
            Case Begin
                Case (iSelectedSize = 0)
                    Function_Return 1000
                Case (iSelectedSize = 1)
                    Function_Return 10000
                Case (iSelectedSize = 2)
                    Function_Return 50000
            Case End
        End_Function  // SizeOfTest
    
        Set Current_Radio to 0
        Set pbFirstTime to False
    End_Object

    Object oTestGroup1 is a Group
        Set Size to 64 185
        Set Location to 5 89
        Set Label to "First Method:"

        Object oTextBox2 is a TextBox
            Set Auto_Size_State to False
            Set Size to 24 173
            Set Location to 12 6
            Set Label to "This method does not use any cache.  The set of names is received and each name is searched  every time."
            Set Justification_Mode to JMode_Left
        End_Object

        Object oRunTest1_btn is a Button
            Set Location to 41 6
            Set Label to "Run Test"
        
            // fires when the button is clicked
            Procedure OnClick
                Integer iTime iCount iSize iLatency
                String[] sNames
                tNameData[] NameData
                
                
                // Get the set of names that we need to lookup....
                Get GetNames to sNames
                
                // Do the test...
                Send Cursor_Wait to Cursor_Control
                Get LookupNamesTest1 sNames (&iTime) to NameData   // we don't do anything to the NameData
                
                // Report the test....
                Get Value of oTestCount1_frm to iCount
                Increment iCount
                Get SizeOfTest of oSizeOfTest_grp to iSize
                Get Value of oLatency_frm to iLatency
                Send ReportTime to oTreeView1 ("First Method (" + String(iCount) + "), " + String(iSize) + ;
                                               " names - Latency = " + String(iLatency) + " ms per name: Total time = ") iTime
                
                Set Value of oTestCount1_frm to iCount
                Send Cursor_Ready to Cursor_Control
            End_Procedure
        
        End_Object

        Object oTestCount1_frm is a Form
            Set Size to 13 40
            Set Location to 42 114
            Set Enabled_State to False
            Set Label to "Test Count:"
            Set Label_Col_Offset to 1
            Set Label_Justification_Mode to JMode_Right
            Set Value to 0
            Set Form_Datatype to 0
        End_Object
    End_Object

    Object oTestGroup2 is a Group
        Set Size to 64 185
        Set Location to 76 89
        Set Label to "Second Method:"

        Object oTextBox2 is a TextBox
            Set Auto_Size_State to False
            Set Size to 24 173
            Set Location to 12 6
            Set Label to "This method uses a cache of all names looked up so far. The cach lookup is searched via a FOR loop."
            Set Justification_Mode to JMode_Left
        End_Object

        Object oRunTest2_btn is a Button
            Set Location to 41 6
            Set Label to "Run Test"
        
            // fires when the button is clicked
            Procedure OnClick
                Integer iTime iCount iSize iLatency
                String[] sNames
                tNameData[] NameData EmptyNameData
                
                // If this is the first test search then ensure that the Cache is cleared....
                Get Value of oTestCount2_frm to iCount
                If (iCount = 0) Begin
                    Set pNameDataCache to EmptyNameData
                End

                // Get the set of names that we need to lookup....
                Get GetNames to sNames
                
                // Do the test...
                Send Cursor_Wait to Cursor_Control
                Get LookupNamesTest2 sNames (&iTime) to NameData   // we don't do anything to the NameData
                
                // Report the test....
                Increment iCount
                Get SizeOfTest of oSizeOfTest_grp to iSize
                Get Value of oLatency_frm to iLatency
                Send ReportTime to oTreeView1 ("Second Method (" + String(iCount) + "), " + String(iSize) + ;
                                               " names - Latency = " + String(iLatency) + " ms per name: Total time = ") iTime
                
                Set Value of oTestCount2_frm to iCount
                Send Cursor_Ready to Cursor_Control
            End_Procedure
        
        End_Object

        Object oTestCount2_frm is a Form
            Set Size to 13 40
            Set Location to 42 114
            Set Enabled_State to False
            Set Label to "Test Count:"
            Set Label_Col_Offset to 1
            Set Label_Justification_Mode to JMode_Right
            Set Value to 0
            Set Form_Datatype to 0
        End_Object
    End_Object

    Object oTestGroup3 is a Group
        Set Size to 64 185
        Set Location to 144 89
        Set Label to "Third Method:"

        Object oTextBox3 is a TextBox
            Set Auto_Size_State to False
            Set Size to 24 173
            Set Location to 12 6
            Set Label to "This method uses a cache of all names looked up so far. The cache lookup is searched via a VDF code SearchArray function."
            Set Justification_Mode to JMode_Left
        End_Object

        Object oRunTest3_btn is a Button
            Set Location to 41 6
            Set Label to "Run Test"
        
            // fires when the button is clicked
            Procedure OnClick
                Integer iTime iCount iSize iLatency
                String[] sNames
                tNameData[] NameData EmptyNameData
                
                // Get the set of names that we need to lookup....
                Get GetNames to sNames
                
                // If this is the first test search then ensure that the Cache is cleared....
                Get Value of oTestCount3_frm to iCount
                If (iCount = 0) Begin
                    Set pNameDataCache to EmptyNameData
                End
                
                // Do the test...
                Send Cursor_Wait to Cursor_Control
                Get LookupNamesTest3 sNames (&iTime) to NameData   // we don't do anything to the NameData
                
                // Report the test....
                Increment iCount
                Get SizeOfTest of oSizeOfTest_grp to iSize
                Get Value of oLatency_frm to iLatency
                Send ReportTime to oTreeView1 ("Third Method (" + String(iCount) + "), " + String(iSize) + ;
                                               " names - Latency = " + String(iLatency) + " ms per name: Total time = ") iTime
                
                Set Value of oTestCount3_frm to iCount
                Send Cursor_Ready to Cursor_Control
            End_Procedure
        
        End_Object

        Object oTestCount3_frm is a Form
            Set Size to 13 40
            Set Location to 42 114
            Set Enabled_State to False
            Set Label to "Test Count:"
            Set Label_Col_Offset to 1
            Set Label_Justification_Mode to JMode_Right
            Set Value to 0
            Set Form_Datatype to 0
        End_Object
    End_Object

    Object oTestGroup4 is a Group
        Set Size to 64 185
        Set Location to 214 89
        Set Label to "Fourth Method:"

        Object oTextBox4 is a TextBox
            Set Auto_Size_State to False
            Set Size to 24 173
            Set Location to 12 6
            Set Label to "This method uses a cache of all names looked up so far. The cache lookup is searched via a runtime SearchArray function."
            Set Justification_Mode to JMode_Left
        End_Object

        Object oRunTest4_btn is a Button
            Set Location to 41 6
            Set Label to "Run Test"
        
            // fires when the button is clicked
            Procedure OnClick
                Integer iTime iCount iSize iLatency
                String[] sNames sEmptyNames
                Date[] dEmptyDates
                tNameData[] NameData
                
                // Get the set of names that we need to lookup....
                Get GetNames to sNames
                
                // If this is the first test search then ensure that the Cache is cleared....
                Get Value of oTestCount4_frm to iCount
                If (iCount = 0) Begin
                    Set pNamesCache to sEmptyNames
                    Set pDatesCache to dEmptyDates
                End
                
                // Do the test...
                Send Cursor_Wait to Cursor_Control
                Get LookupNamesTest4 sNames (&iTime) to NameData   // we don't do anything to the NameData
                
                // Report the test....
                Increment iCount
                Get SizeOfTest of oSizeOfTest_grp to iSize
                Get Value of oLatency_frm to iLatency
                Send ReportTime to oTreeView1 ("Fourth Method (" + String(iCount) + "), " + String(iSize) + ;
                                               " names - Latency = " + String(iLatency) + " ms per name: Total time = ") iTime
                
                Set Value of oTestCount4_frm to iCount
                Send Cursor_Ready to Cursor_Control
            End_Procedure
        
        End_Object

        Object oTestCount4_frm is a Form
            Set Size to 13 40
            Set Location to 42 114
            Set Enabled_State to False
            Set Label to "Test Count:"
            Set Label_Col_Offset to 1
            Set Label_Justification_Mode to JMode_Right
            Set Value to 0
            Set Form_Datatype to 0
        End_Object
    End_Object

    Object oTestGroup5 is a Group
        Set Size to 85 185
        Set Location to 285 89
        Set Label to "Fifth Method:"

        Object oTextBox5 is a TextBox
            Set Auto_Size_State to False
            Set Size to 51 173
            Set Location to 12 6
            Set Label to "This method uses two caches of all names looked up so far: a sorted cache and an unsorted cache. The two caches are mutually exclusive (no duplicates). The sorted cache is searched via BinarySearchArray. The unsorted cache is searched via a runtime SearchArray function."
            Set Justification_Mode to JMode_Left
        End_Object

        Object oRunTest5_btn is a Button
            Set Location to 66 6
            Set Label to "Run Test"
        
            // fires when the button is clicked
            Procedure OnClick
                Integer iTime iCount iSize iLatency
                String[] sNames sEmptyNames
                Date[] dEmptyDates
                tNameData[] NameData
                
                // Get the set of names that we need to lookup....
                Get GetNames to sNames
                
                // If this is the first test search then ensure that the Cache is cleared....
                Get Value of oTestCount5_frm to iCount
                If (iCount = 0) Begin
                    Set pNamesCache to sEmptyNames
                    Set pDatesCache to dEmptyDates
                    Set pSortedNamesCache to sEmptyNames
                    Set pSortedDatesCache to dEmptyDates
                End
                
                // Do the test...
                Send Cursor_Wait to Cursor_Control
                Get LookupNamesTest5 sNames (&iTime) to NameData   // we don't do anything to the NameData
                
                // Report the test....
                Increment iCount
                Get SizeOfTest of oSizeOfTest_grp to iSize
                Get Value of oLatency_frm to iLatency
                Send ReportTime to oTreeView1 ("Fifth Method (" + String(iCount) + "), " + String(iSize) + ;
                                               " names - Latency = " + String(iLatency) + " ms per name: Total time = ") iTime
                
                Set Value of oTestCount5_frm to iCount
                Send Cursor_Ready to Cursor_Control
            End_Procedure
        
        End_Object

        Object oTestCount5_frm is a Form
            Set Size to 13 40
            Set Location to 67 114
            Set Enabled_State to False
            Set Label to "Test Count:"
            Set Label_Col_Offset to 1
            Set Label_Justification_Mode to JMode_Right
            Set Value to 0
            Set Form_Datatype to 0
        End_Object
    End_Object

    Object oLatency_grp is a Group
        Set Size to 43 77
        Set Location to 76 6
        Set Label to "Latency:"

        Object oLatency_frm is a SpinForm
            Set Size to 13 63
            Set Location to 25 7
            Set Label to "milliseconds/lookup"
            Set Label_Col_Offset to 0
            Set Label_Justification_Mode to JMode_Top
            Set Value to "5"
            Set Form_Datatype to 0
        End_Object
    End_Object

    Object oTextBox1 is a TextBox
        Set Size to 10 27
        Set Location to 6 281
        Set Label to "Results:"
    End_Object

    Object oTreeView1 is a TreeView
        Set Size to 351 223
        Set Location to 18 281
        Set Label to "Results:"
        Set TreeButtonsState to False
        Set TreeLinesState to False
        Set TreeRootLinesState to False
        Set peAnchors to anAll
        
        // ReportTime:
        
        // Adds a new root item to the treeview item list
        // iTime - Is the time in milliseconds.
        Procedure ReportTime String sLabel Integer iTime
            Handle hItem
            String sUnits
            Number nTime
            
            Move " ms" to sUnits    // the default
            
            If (iTime > 1000) Begin
                // Convert milliseconds into seconds....
                Move (iTime / 1000.0) to nTime
                Move " sec" to sUnits 
            End
            Else Move iTime to nTime
            
            Get AddTreeItem (sLabel + " " + (String(nTime)) + sUnits) 0 0 0 0 to hItem
        End_Procedure  // ReportTime
    End_Object  // oTreeView1

    Object oTestLoopOptimiztion_btn is a Button
        Set Size to 14 78
        Set Location to 124 6
        Set Label to "Test Loop Optimization"
    
        // This tests the difference between performing a FOR loop where the
        // stop value is decremented for each loop against a FOR loop where the
        // stop value is decremented once before the looping begins.
        Procedure OnClick
            Integer iSizeOfTest i
            Integer iStart iStop iTime
            
            Get SizeOfTest of oSizeOfTest_grp to iSizeOfTest
            
            // Perform the first test....
            Move (GetTickCount()) to iStart
            
            For i from 0 to (iSizeOfTest - 1)
            Loop
            
            Move (GetTickCount()) to iStop
            Move (iStop - iStart) to iTime
            Send ReportTime to oTreeView1 ("Unoptimized loop test (" + String(iSizeOfTest) + ") loops") iTime
            
            // Perform the second test....
            Move (GetTickCount()) to iStart
            Decrement iSizeOfTest
            
            For i from 0 to iSizeOfTest
            Loop
            
            Move (GetTickCount()) to iStop
            Move (iStop - iStart) to iTime
            Send ReportTime to oTreeView1 ("Optimized loop test (" + String(iSizeOfTest+1) + ") loops") iTime
        End_Procedure
    End_Object

    
    // CompareNameData:
    
    // Custom comparison function for SearchArray of tNameData types.
    // Returns (EQ) if the sName members of the passed parameters match.
    Function CompareNameData tNameData SearchName tNameData ReferenceName Returns Integer
        If (SearchName.sName = ReferenceName.sName) Function_Return (EQ)
        Else Function_Return (NE)
    End_Function  // CompareNameData
    
    
    // GetNames:
    
    // Simulates getting a set of names. This method randomly generates a set of 10-character names.
    // We use these names in our test. Generating the names is not part of the test so we don't care that
    // it is inefficient.
    Function GetNames Returns String[]
        Integer icTestData iTotalNames iTestDataOffset
        Boolean bStoreNames
        String[] sNames sTestData
        
        // Does the total set of names need to be generated? This is done once only before the first test....
        Get pTestData to sTestData
        Move (SizeOfArray(sTestData)) to icTestData
        Move False to bStoreNames

        If (icTestData = 0) Begin
            // First we'll try to read in the set of all names from an external text file....
            Move True to bStoreNames
            Send ReadTestDataFromTextFile (&sTestData)
            Move (SizeOfArray(sTestData)) to icTestData
        End
       
        If (icTestData < C_SampleSize) Begin
            // If the text fle doesn't exist we'll generate our test data and then write it out to the text file....
            Send GenerateTestData (&sTestData)
            Send WriteTestDataToTextfile sTestData
        End

        // If we had to read or generate the names then store them in a property for next time....        
        If (bStoreNames) Begin
            Set pTestData to sTestData
        End
        

        // Get the selected size of the test. i.e. the number of names to be returned....
        Get SizeOfTest of oSizeOfTest_grp to iTotalNames
        Get piTestDataOffset to iTestDataOffset               // iTestDataOffset is incremented each time to ensure we fetch some new records from the test data
        
        If ((iTotalNames + iTestDataOffset) > C_SampleSize) Begin
            Move 0 to iTestDataOffset
        End
        
        // Fetch the set of names out of the test data....
        Move (CopyArray(sTestData, iTestDataOffset, iTotalNames + iTestDataOffset - 1)) to sNames
        Set piTestDataOffset to (iTestDataOffset + C_TestDataIncrementSize)
        
        Function_Return sNames
    End_Procedure  // GetNames


    // GenerateTestData:
    
    // Generates the set of test data that we will use. The test data is an array of 110,000 randomly generated names.
    Procedure GenerateTestData String[] ByRef sNames
        Integer iCount icName iDuplicates iVal iFound
        String sName sSampleSize
        Boolean bUserCancel

        Send Initialize_StatusPanel of ghoStatusPanel "Generating Test Data" "Sorry, this takes a long time but it is only needed once" "" ""
        Send Start_StatusPanel of ghoStatusPanel

        Move (SizeOfArray(sNames)) to icName
        Move (String(C_SampleSize)) to sSampleSize
        Move 0 to iDuplicates
        
        // Generate the set of random names....
        While (icName < C_SampleSize)
            Move "" to sName
            // Each name is 6-characters long....
            For iCount from 0 to 4
                Move (Random(25)) to iVal    // random number between 0-25
                Move (iVal + 65) to iVal     // convert number to ASCII value of an uppercase letter
                Move (sName + Character(iVal)) to sName  // Append the character to the name we are generating
            Loop
            
            // Does the name already belong to the list?....
            Move (SearchArray(sName, sNames)) to iFound
            
            If (iFound = -1) Begin
                // Add the new name to the list....
                Move sName to sNames[icName]
                Increment icName
                Send Update_StatusPanel of ghoStatusPanel (String(icName) + " of " + sSampleSize)
            End
            Else Begin
                Increment iDuplicates
                Set Message_Text of ghoStatusPanel to ("Duplicates Skipped: " + String(iDuplicates))
            End
            
            // Test for operator cancel....
            Get Check_StatusPanel of ghoStatusPanel to bUserCancel
            If (bUserCancel) Begin
                If (YesNo_Box("Do you want to quit this application?", "Confirm Abort Application") = MBR_Yes) Begin
                    Abort
                End
            End
        Loop

        Send Stop_StatusPanel of ghoStatusPanel
    End_Procedure  // GenerateTestData


    // GetDateOfBirth:
    
    // Simulates looking up the passed name by calling a web service to get the date of births for that name.
    Function GetDateOfBirth Returns Date
        Integer iTestLatency iVoid iDay iMonth iYear
        
        // simulate waiting for the specified time....
        Get Value of oLatency_frm to iTestLatency
        Move (Sleep(iTestLatency)) to iVoid
        Move (Random(11)) to iDay
        Move (Random(11)) to iMonth
        Move (Random(9)) to iYear
        Increment iDay
        Increment iMonth
        Move (iYear + 2000) to iYear
        
        Function_Return (Date(String(iDay) + "/" + String(iMonth) + "/" + String(iYear)))
        Function_Return 01/01/2009
    End_Function  // GetDateOfBirth
    

    // LookupNamesTest1:
    
    // Performs a lookup for each name of the passed array of names using the first method.
    // The lookup simulates getting some data for the passed name via a web service.
    // This method is simple, the lookup is performed on each name.
    // sNames - Is the array of names to be looked up.
    // iTime  - [out] is the time taken to complete the test in (ms).
    // Returns - The set of names and sales volume figures.
    Function LookupNamesTest1 String[] sNames Integer ByRef iTime Returns tNameData[]
        Integer iStart iStop iVoid
        Integer iName icName
        tNameData[] NameData
        
        Move (TimeBeginPeriod(1)) to iVoid
        Move (GetTickCount()) to iStart
        
        Move (SizeOfArray(sNames)) to icName
        Decrement icName
        
        For iName from 0 to icName
            Move sNames[iName]               to NameData[iName].sName
            Get GetDateOfBirth sNames[iName] to NameData[iName].dDateOfBirth
        Loop
        
        Move (GetTickCount()) to iStop
        Move (iStop - iStart) to iTime
        Move (TimeEndPeriod(1)) to iVoid
        Function_Return NameData
    End_Function  // LookupNamesTest1
    

    // LookupNamesTest2:
    
    // Performs a lookup for each name of the passed array of names using the first method.
    // The lookup simulates getting some data for the passed name via a web service.
    // This method uses a cache of all names looked up so far. Only names that are not in the
    // cache are looked up. However searching the cache is not very efficient. We use a FOR loop
    // to find each name in the cache.
    // sNames - Is the array of names to be looked up.
    // iTime  - [out] is the time taken to complete the test in (ms).
    // Returns - The set of names and sales volume figures.
    Function LookupNamesTest2 String[] sNames Integer ByRef iTime Returns tNameData[]
        Integer iStart iStop
        Integer iName icName
        Integer iCache icCache iTopOfCache iFound iVoid
        Boolean bFound bChanged
        tNameData[] NameData NameCacheOld NameCacheNew
        
        Move (TimeBeginPeriod(1)) to iVoid
        Move (GetTickCount()) to iStart
        
        // Get the cached data....
        Move False to bChanged
        Get pNameDataCache to NameCacheOld
        Move NameCacheOld to NameCacheNew           // make a copy of the cache
        Move (SizeOfArray(NameCacheOld)) to icCache
        Move icCache to iTopOfCache
        Decrement icCache
        
        // Iterate the set of names....
        Move (SizeOfArray(sNames)) to icName
        Decrement icName
        
        For iName from 0 to icName
            // Test if the name belongs to our original set of cached data. We don't lookup
            // the names in the new cached data set because we know the names are unique so
            // if it is not in the original set then it is not in the new set....
            // We just use a For Loop here. This is not very fast.
            Move False to bFound
            
            For iCache from 0 to icCache
                Move (sNames[iName] = NameCacheOld[iCache].sName) to bFound
                
                If (bFound) Begin
                    Move iCache to iFound
                End
            
                If (bFound) Break    // stop looking
            Loop
            
            Move sNames[iName] to NameData[iName].sName
            
            If (bFound) Begin
                // If found, then use the cached value....
                Move NameCacheOld[iFound].dDateOfBirth to NameData[iName].dDateOfBirth
            End
            Else Begin
                // Otherwise we have to look it up and add the value to the cache....
                Get GetDateOfBirth sNames[iName] to NameData[iName].dDateOfBirth
                
                Move NameData[iName].sName        to NameCacheNew[iTopOfCache].sName
                Move NameData[iName].dDateOfBirth to NameCacheNew[iTopOfCache].dDateOfBirth
                Increment iTopOfCache
                Move True to bChanged
            End
        Loop
        
        // If the cached has been updated then we need to save the changes....
        If (bChanged) Begin
            Set pNameDataCache to NameCacheNew
        End
        
        Move (GetTickCount()) to iStop
        Move (iStop - iStart) to iTime
        Move (TimeEndPeriod(1)) to iVoid
        Function_Return NameData
    End_Function  // LookupNamesTest2


    // LookupNamesTest3:
    
    // Performs a lookup for each name of the passed array of names using the first method.
    // The lookup simulates getting some data for the passed name via a web service.
    // This method uses a cache of all names looked up so far. Only names that are not in the
    // cache are looked up. However searching the cache is still not very efficient. We use
    // SearchArray via a VDF comparison function to handle tNameData struct arrays.
    // sNames - Is the array of names to be looked up.
    // iTime  - [out] is the time taken to complete the test in (ms).
    // Returns - The set of names and sales volume figures.
    Function LookupNamesTest3 String[] sNames Integer ByRef iTime Returns tNameData[]
        Integer iStart iStop
        Integer iName icName
        Integer icCache iTopOfCache iFound iVoid
        Boolean bChanged
        tNameData SearchName
        tNameData[] NameData NameCacheOld NameCacheNew
        
        Move (TimeBeginPeriod(1)) to iVoid
        Move (GetTickCount()) to iStart
        
        // Get the cached data....
        Move False to bChanged
        Get pNameDataCache to NameCacheOld
        Move NameCacheOld to NameCacheNew           // make a copy of the cache
        Move (SizeOfArray(NameCacheOld)) to icCache
        Move icCache to iTopOfCache
        
        // Iterate the set of names....
        Move (SizeOfArray(sNames)) to icName
        Decrement icName
        
        For iName from 0 to icName
            // Test if the name belongs to our original set of cached data. We don't lookup
            // the names in the new cached data set because we know the names are unique so
            // if it is not in the original set then it is not in the new set....
            Move -1 to iFound
            
            If (icCache > 0) Begin    // don't try to search the cache if it is empty. This speeds up the first search.
                Move sNames[iName] to SearchName.sName
                Move (SearchArray(SearchName, NameCacheOld, Self, get_CompareNameData)) to iFound
            End
            
            Move sNames[iName] to NameData[iName].sName
            
            If (iFound >= 0) Begin
                // If found, then use the cached value....
                Move NameCacheOld[iFound].dDateOfBirth to NameData[iName].dDateOfBirth
            End
            Else Begin
                // Otherwise we have to look it up and add the value to the cache....
                Get GetDateOfBirth sNames[iName] to NameData[iName].dDateOfBirth
                
                Move NameData[iName].sName        to NameCacheNew[iTopOfCache].sName
                Move NameData[iName].dDateOfBirth to NameCacheNew[iTopOfCache].dDateOfBirth
                Increment iTopOfCache
                Move True to bChanged
            End
        Loop
        
        // If the cached has been updated then we need to save the changes....
        If (bChanged) Begin
            Set pNameDataCache to NameCacheNew
        End
        
        Move (GetTickCount()) to iStop
        Move (iStop - iStart) to iTime
        Move (TimeEndPeriod(1)) to iVoid
        Function_Return NameData
    End_Function  // LookupNamesTest3


    // LookupNamesTest4:
    
    // Performs a lookup for each name of the passed array of names using the first method.
    // The lookup simulates getting some data for the passed name via a web service.
    // This method uses a cache of all names looked up so far. Only names that are not in the
    // cache are looked up.
    //
    // We Use SearchArray via a runtime comparison function. To perform this search we must split the
    // cache into two separate synchronized arrays. One for each member of tNameData.
    //
    // sNames - Is the array of names to be looked up.
    // iTime  - [out] is the time taken to complete the test in (ms).
    // Returns - The set of names and sales volume figures.
    Function LookupNamesTest4 String[] sNames Integer ByRef iTime Returns tNameData[]
        Integer iStart iStop
        Integer iName icName
        Integer icCache iTopOfCache iFound iVoid
        Boolean bChanged
        String[] NamesCacheOld NamesCacheNew
        Date[] DatesCacheOld DatesCacheNew
        tNameData[] NameData
        
        Move (TimeBeginPeriod(1)) to iVoid
        Move (GetTickCount()) to iStart
        
        // Get the cached data....
        Move False to bChanged
        Get pNamesCache to NamesCacheOld
        Get pDatesCache to DatesCacheOld
        Move NamesCacheOld to NamesCacheNew           // make a copy of the cache
        Move DatesCacheOld to DatesCacheNew
        Move (SizeOfArray(NamesCacheOld)) to icCache
        Move icCache to iTopOfCache
        
        // Iterate the set of names....
        Move (SizeOfArray(sNames)) to icName
        Decrement icName
        
        For iName from 0 to icName
            // Test if the name belongs to our original set of cached data. We don't lookup
            // the names in the new cached data set because we know the names are unique so
            // if it is not in the original set then it is not in the new set....
            Move -1 to iFound
            
            If (icCache > 0) Begin    // don't try to search the cache if it is empty. This speeds up the first search.
                Move (SearchArray(sNames[iName], NamesCacheOld)) to iFound
            End
            
            Move sNames[iName] to NameData[iName].sName
            
            If (iFound >= 0) Begin
                // If found, then use the cached value....
                Move DatesCacheOld[iFound] to NameData[iName].dDateOfBirth
            End
            Else Begin
                // Otherwise we have to look it up and add the value to the cache....
                Get GetDateOfBirth sNames[iName] to NameData[iName].dDateOfBirth
                
                Move NameData[iName].sName        to NamesCacheNew[iTopOfCache]
                Move NameData[iName].dDateOfBirth to DatesCacheNew[iTopOfCache]
                Increment iTopOfCache
                Move True to bChanged
            End
        Loop
        
        // If the cached has been updated then we need to save the changes....
        If (bChanged) Begin
            Set pNamesCache to NamesCacheNew
            Set pDatesCache to DatesCacheNew
        End
        
        Move (GetTickCount()) to iStop
        Move (iStop - iStart) to iTime
        Move (TimeEndPeriod(1)) to iVoid
        Function_Return NameData
    End_Function  // LookupNamesTest4

    
    // LookupNamesTest5:
    
    // Performs a lookup for each name of the passed array of names using the first method.
    // The lookup simulates getting some data for the passed name via a web service.
    // This method uses a cache of all names looked up so far. Only names that are not in the
    // cache are looked up. 
    //
    // If the cache is empty then we will sort the resulted data set and store it in a sorted cache.
    // Once the sorted cache is established then any additional names in subsequent lookups are stored
    // in an unsorted cache.
    // 
    // When we lookup a set of names then the sorted cache is used first using BinarySearchArray. If the
    // name was not found there then the unsorted array is used via SearchArray. If the name is still not 
    // Found then we cll GetDateOfBirth to get the date.    
    //
    // sNames - Is the array of names to be looked up.
    // iTime  - [out] is the time taken to complete the test in (ms).
    // Returns - The set of names and sales volume figures.
    Function LookupNamesTest5 String[] sNames Integer ByRef iTime Returns tNameData[]
        Integer iStart iStop
        Integer iName icName
        Integer icSortedCache icUnsortedCache iTopOfUnsortedCache iFound iVoid
        Integer i iCount
        Boolean bChanged
        String[] NamesCacheOld NamesCacheNew SortedNamesCache
        Date[] DatesCacheOld DatesCacheNew SortedDatesCache
        tNameData[] NameData
        
        Move (TimeBeginPeriod(1)) to iVoid
        Move (GetTickCount()) to iStart
        
        // Get the cached data....
        Move False to bChanged
        Get pNamesCache to NamesCacheOld
        Get pDatesCache to DatesCacheOld
        Move NamesCacheOld to NamesCacheNew           // make a copy of the cache
        Move DatesCacheOld to DatesCacheNew
        Move (SizeOfArray(NamesCacheOld)) to icUnsortedCache
        Move icUnsortedCache to iTopOfUnsortedCache
    
        Get pSortedNamesCache to SortedNamesCache
        Get pSortedDatesCache to SortedDatesCache
        Move (SizeOfArray(SortedNamesCache)) to icSortedCache
        
        // Iterate the set of names....
        Move (SizeOfArray(sNames)) to icName
        Decrement icName
        
        For iName from 0 to icName
            Move sNames[iName] to NameData[iName].sName
            
            // First search the sorted cache....
            Move -1 to iFound
            
            If (icSortedCache > 0) Begin    // don't try to search the cache if it is empty. This speeds up the first search.
                Move (BinarySearchArray(sNames[iName], SortedNamesCache)) to iFound
            End
            
            // If found, then use the cached value....
            If (iFound >= 0) Begin
                Move SortedDatesCache[iFound] to NameData[iName].dDateOfBirth
            End
            Else Begin
                // if we didn't find anything then search the unsorted cache....
                If (icUnsortedCache > 0) Begin    // don't try to search the cache if it is empty. This speeds up the first search.
                    Move (SearchArray(sNames[iName], NamesCacheOld)) to iFound
                End
    
                // If found, then use the cached value....
                If (iFound >= 0) Begin
                    Move DatesCacheOld[iFound] to NameData[iName].dDateOfBirth
                End
            End
            
            // If not found, then we have to look it up and add the value to the cache....
            If (iFound < 0) Begin
                Get GetDateOfBirth sNames[iName] to NameData[iName].dDateOfBirth
                
                Move NameData[iName].sName        to NamesCacheNew[iTopOfUnsortedCache]
                Move NameData[iName].dDateOfBirth to DatesCacheNew[iTopOfUnsortedCache]
                Increment iTopOfUnsortedCache
                Move True to bChanged
            End
        Loop
        
        // If the cached has been updated then we need to save the changes....
        If (bChanged) Begin
            // If the sorted cache is empty then we will initialize it with the data we have now....
            If (icSortedCache = 0) Begin
                // First sort the names....
                Move (SortArray(NamesCacheNew)) to SortedNamesCache
                
                // Now lookup each name in the Sorted list to create the corresponding sorted dates....
                Move (iTopOfUnsortedCache - 1) to iCount
                
                For i from 0 to iCount
                    Move (BinarySearchArray(NamesCacheNew[i], SortedNamesCache)) to iFound    // get the index into the sorted names array
                    Move (DatesCacheNew[i]) to SortedDatesCache[iFound]    // Save the corresponding date in the sorted position
                Loop
                
                // Save the sorted cache data....
                Set pSortedNamesCache to SortedNamesCache
                Set pSortedDatesCache to SortedDatesCache
            End
            Else Begin
                // Otherwise we just update the unsorted cache....
                Set pNamesCache to NamesCacheNew
                Set pDatesCache to DatesCacheNew
            End
        End
        
        Move (GetTickCount()) to iStop
        Move (iStop - iStart) to iTime
        Move (TimeEndPeriod(1)) to iVoid
        Function_Return NameData
    End_Function  // LookupNamesTest5

    
    // ReadTestDataFromTextFile:
    
    // Reads in our test data from a text file and returns them in a string array.
    Procedure ReadTestDataFromTextFile String[] ByRef sNames
        Integer iCh iCount
        Integer iStart iStop
        String sName sFileName sSep sSampleSize
        Boolean bUserCancel

        Move (GetTickCount()) to iStart
        Send Initialize_StatusPanel of ghoStatusPanel "Reading Test Data" "" "" ""
        Send Start_StatusPanel of ghoStatusPanel

        Move (SysConf(SYSCONF_DIR_SEPARATOR)) to sSep
        Get GetApplicationPath of ghoApplication to sFileName
        Move (sFileName + sSep + C_TestDataFileName) to sFileName
        Move 0 to iCount
        Move (Seq_New_Channel()) to iCh
        Move (String(C_SampleSize)) to sSampleSize
//        Move (ResizeArray(sNames, C_SampleSize)) to sNames   <-- This does not make it faster!
        
        Direct_Input channel iCh sFileName
        
        Repeat
            Readln channel iCh sName
            
            If (not(SeqEof)) Begin
                Move sName to sNames[iCount]
                Increment iCount
                Send Update_StatusPanel of ghoStatusPanel (String(iCount) + " of " + sSampleSize)
            End
                
            // Test for operator cancel....
            Get Check_StatusPanel of ghoStatusPanel to bUserCancel
            If (bUserCancel) Begin
                If (YesNo_Box("Do you want to quit this application?", "Confirm Abort Application") = MBR_Yes) Begin
                    Abort
                End
            End
        Until (SeqEof)

        Send Stop_StatusPanel of ghoStatusPanel
        
        Close_Input channel iCh
        
        Send Seq_Release_Channel iCh

        Move (GetTickCount()) to iStop
        Send ReportTime to oTreeView1 "Read test data: " (iStop - iStart)
    End_Procedure  // ReadTestDataFromTextFile


    // WriteTestDataToTextfile:
    
    // Writes the test data out to a tex file. That saves us time the next time we run this test since we don't
    // have to re-generate the names which takes a long time.
    Procedure WriteTestDataToTextfile String[] sNames
        Integer iCh iCount
        String sName sFileName sSep sSampleSize
        Boolean bUserCancel

        Send Initialize_StatusPanel of ghoStatusPanel "Writing Test Data" "" "" ""
        Send Start_StatusPanel of ghoStatusPanel

        Move (SysConf(SYSCONF_DIR_SEPARATOR)) to sSep
        Get GetApplicationPath of ghoApplication to sFileName
        Move (sFileName + sSep + C_TestDataFileName) to sFileName
        Move 0 to iCount
        Move (Seq_New_Channel()) to iCh
        Move (String(C_SampleSize)) to sSampleSize
        
        Direct_Output channel iCh sFileName
        
        For iCount from 0 to (C_SampleSize - 1)
            Writeln channel iCh sNames[iCount]
            Send Update_StatusPanel of ghoStatusPanel (String(iCount) + " of " + sSampleSize)
                
            // Test for operator cancel....
            Get Check_StatusPanel of ghoStatusPanel to bUserCancel
            If (bUserCancel) Begin
                If (YesNo_Box("Do you want to quit this application?", "Confirm Abort Application") = MBR_Yes) Begin
                    Abort
                End
            End
        Until (SeqEof)

        Send Stop_StatusPanel of ghoStatusPanel
        
        Close_Output channel iCh
        
        Send Seq_Release_Channel iCh
    End_Procedure  // WriteTestDataToTextfile    
End_Object
